@@ -9,6 +9,7 @@
 #include <stdbool.h>
 
 #include "core/list.h"
+#include "core/runtime.h"
 
 struct bf_chain;
 struct bf_hookopts;
@@ -35,3 +36,13 @@ void bfc_chain_dump(struct bf_chain *chain, struct bf_hookopts *hookopts,
  * @param counters List of counters to print.
  */
 int bfc_ruleset_dump(bf_list *chains, bf_list *hookopts, bf_list *counters);
+
+/**
+ * @brief Print a logged packet published by a rule.
+ *
+ * @pre
+ * - `log != NULL`
+ *
+ * @param log Logged data.
+ */
+void bfc_print_log(const struct bf_log *log);
