@@ -23,6 +23,7 @@
 #include "bpfilter/cgen/dump.h"
 #include "bpfilter/cgen/fixup.h"
 #include "bpfilter/cgen/jmp.h"
+#include "bpfilter/cgen/matcher/icmp.h"
 #include "bpfilter/cgen/matcher/ip4.h"
 #include "bpfilter/cgen/matcher/ip6.h"
 #include "bpfilter/cgen/matcher/meta.h"
@@ -60,6 +61,20 @@
 #define _BF_LOG_BUF_SIZE                                                       \
     (UINT32_MAX >> 8) /* verifier maximum in kernels <= 5.1 */
 #define _BF_PROGRAM_DEFAULT_IMG_SIZE (1 << 6)
+#define _BF_LOG_MAP_N_ENTRIES 1000
+#define _BF_LOG_MAP_SIZE _bf_round_next_power_of_2(sizeof(struct bf_log) * _BF_LOG_MAP_N_ENTRIES)
+
+static inline size_t _bf_round_next_power_of_2(size_t value)
+{
+    value--;
+    value |= value >> 1;
+    value |= value >> 2;
+    value |= value >> 4;
+    value |= value >> 8;
+    value |= value >> 16;
+
+    return ++value;
+}
 
 static const struct bf_flavor_ops *bf_flavor_ops_get(enum bf_flavor flavor)
 {
@@ -78,7 +93,7 @@ static const struct bf_flavor_ops *bf_flavor_ops_get(enum bf_flavor flavor)
 
 int bf_program_new(struct bf_program **program, const struct bf_chain *chain)
 {
-    _cleanup_bf_program_ struct bf_program *_program = NULL;
+    _free_bf_program_ struct bf_program *_program = NULL;
     char name[BPF_OBJ_NAME_LEN];
     uint32_t set_idx = 0;
     int r;
@@ -108,14 +123,20 @@ int bf_program_new(struct bf_program **program, const struct bf_chain *chain)
     if (r < 0)
         return bf_err_r(r, "failed to create the printer bf_map object");
 
+    r = bf_map_new(&_program->lmap, "log_map", BF_MAP_TYPE_LOG,
+                   BF_MAP_BPF_TYPE_RINGBUF, 0, 0, _BF_LOG_MAP_SIZE);
+    if (r < 0)
+        return bf_err_r(r, "failed to create the log bf_map object");
+
     _program->sets = bf_map_list();
     bf_list_foreach (&chain->sets, set_node) {
         struct bf_set *set = bf_list_node_get_data(set_node);
-        _cleanup_bf_map_ struct bf_map *map = NULL;
+        _free_bf_map_ struct bf_map *map = NULL;
 
         (void)snprintf(name, BPF_OBJ_NAME_LEN, "set_%04x", (uint8_t)set_idx++);
-        r = bf_map_new(&map, name, BF_MAP_TYPE_SET, BF_MAP_BPF_TYPE_HASH,
-                       set->elem_size, 1, bf_list_size(&set->elems));
+        r = bf_map_new(&map, name, BF_MAP_TYPE_SET,
+                       bf_map_bpf_type_from_set_type(set->type), set->elem_size,
+                       1, bf_list_size(&set->elems));
         if (r < 0)
             return r;
 
@@ -158,6 +179,7 @@ void bf_program_free(struct bf_program **program)
 
     bf_map_free(&(*program)->cmap);
     bf_map_free(&(*program)->pmap);
+    bf_map_free(&(*program)->lmap);
     bf_list_clean(&(*program)->sets);
     bf_link_free(&(*program)->link);
     bf_printer_free(&(*program)->printer);
@@ -168,7 +190,7 @@ void bf_program_free(struct bf_program **program)
 
 int bf_program_marsh(const struct bf_program *program, struct bf_marsh **marsh)
 {
-    _cleanup_bf_marsh_ struct bf_marsh *_marsh = NULL;
+    _free_bf_marsh_ struct bf_marsh *_marsh = NULL;
     int r;
 
     bf_assert(program);
@@ -180,7 +202,7 @@ int bf_program_marsh(const struct bf_program *program, struct bf_marsh **marsh)
 
     {
         // Serialize bf_program.counters
-        _cleanup_bf_marsh_ struct bf_marsh *counters_elem = NULL;
+        _free_bf_marsh_ struct bf_marsh *counters_elem = NULL;
 
         r = bf_map_marsh(program->cmap, &counters_elem);
         if (r < 0)
@@ -193,7 +215,7 @@ int bf_program_marsh(const struct bf_program *program, struct bf_marsh **marsh)
 
     {
         // Serialize bf_program.pmap
-        _cleanup_bf_marsh_ struct bf_marsh *pmap_elem = NULL;
+        _free_bf_marsh_ struct bf_marsh *pmap_elem = NULL;
 
         r = bf_map_marsh(program->pmap, &pmap_elem);
         if (r < 0)
@@ -204,9 +226,22 @@ int bf_program_marsh(const struct bf_program *program, struct bf_marsh **marsh)
             return r;
     }
 
+    {
+        // Serialize bf_program.lmap
+        _free_bf_marsh_ struct bf_marsh *lmap_elem = NULL;
+
+        r = bf_map_marsh(program->lmap, &lmap_elem);
+        if (r < 0)
+            return r;
+
+        r = bf_marsh_add_child_obj(&_marsh, lmap_elem);
+        if (r < 0)
+            return r;
+    }
+
     {
         // Serialize bf_program.sets
-        _cleanup_bf_marsh_ struct bf_marsh *sets_elem = NULL;
+        _free_bf_marsh_ struct bf_marsh *sets_elem = NULL;
 
         r = bf_list_marsh(&program->sets, &sets_elem);
         if (r < 0)
@@ -222,7 +257,7 @@ int bf_program_marsh(const struct bf_program *program, struct bf_marsh **marsh)
 
     {
         // Serialize bf_program.links
-        _cleanup_bf_marsh_ struct bf_marsh *links_elem = NULL;
+        _free_bf_marsh_ struct bf_marsh *links_elem = NULL;
 
         r = bf_link_marsh(program->link, &links_elem);
         if (r)
@@ -238,7 +273,7 @@ int bf_program_marsh(const struct bf_program *program, struct bf_marsh **marsh)
 
     {
         // Serialise bf_program.printer
-        _cleanup_bf_marsh_ struct bf_marsh *child = NULL;
+        _free_bf_marsh_ struct bf_marsh *child = NULL;
 
         r = bf_printer_marsh(program->printer, &child);
         if (r)
@@ -249,8 +284,6 @@ int bf_program_marsh(const struct bf_program *program, struct bf_marsh **marsh)
             return bf_err_r(r, "failed to append object to marsh");
     }
 
-    r |= bf_marsh_add_child_raw(&_marsh, &program->num_counters,
-                                sizeof(program->num_counters));
     r |= bf_marsh_add_child_raw(&_marsh, program->img,
                                 program->img_size * sizeof(struct bpf_insn));
     if (r)
@@ -265,7 +298,8 @@ int bf_program_unmarsh(const struct bf_marsh *marsh,
                        struct bf_program **program,
                        const struct bf_chain *chain, int dir_fd)
 {
-    _cleanup_bf_program_ struct bf_program *_program = NULL;
+    _free_bf_program_ struct bf_program *_program = NULL;
+    _free_bf_link_ struct bf_link *link = NULL;
     struct bf_marsh *child = NULL;
     int r;
 
@@ -289,6 +323,13 @@ int bf_program_unmarsh(const struct bf_marsh *marsh,
     if (r < 0)
         return r;
 
+    if (!(child = bf_marsh_next_child(marsh, child)))
+        return -EINVAL;
+    bf_map_free(&_program->lmap);
+    r = bf_map_new_from_marsh(&_program->lmap, dir_fd, child);
+    if (r < 0)
+        return r;
+
     /** @todo Avoid creating and filling the list in @ref bf_program_new before
      * trashing it all here. Eventually, this function will be replaced with
      * @c bf_program_new_from_marsh and this issue could be solved by **not**
@@ -303,7 +344,7 @@ int bf_program_unmarsh(const struct bf_marsh *marsh,
         struct bf_marsh *set_elem = NULL;
 
         while ((set_elem = bf_marsh_next_child(child, set_elem))) {
-            _cleanup_bf_map_ struct bf_map *map = NULL;
+            _free_bf_map_ struct bf_map *map = NULL;
 
             r = bf_map_new_from_marsh(&map, dir_fd, set_elem);
             if (r < 0)
@@ -321,9 +362,13 @@ int bf_program_unmarsh(const struct bf_marsh *marsh,
     if (!(child = bf_marsh_next_child(marsh, child)))
         return -EINVAL;
 
-    bf_link_free(&_program->link);
-    r = bf_link_new_from_marsh(&_program->link, dir_fd, child);
-    if (r)
+    /* Try to restore the link: on success, replace the program's link with the
+     * restored on. If -ENOENT is returned, the link doesn't exist, meaning the
+     * program is not attached. Otherwise, return an error. */
+    r = bf_link_new_from_marsh(&link, dir_fd, child);
+    if (!r)
+        bf_swap(_program->link, link);
+    else if (r != -ENOENT)
         return bf_err_r(r, "failed to restore bf_program.link");
 
     // Unmarsh bf_program.printer
@@ -336,11 +381,6 @@ int bf_program_unmarsh(const struct bf_marsh *marsh,
     if (r)
         return bf_err_r(r, "failed to restore bf_printer object");
 
-    if (!(child = bf_marsh_next_child(marsh, child)))
-        return -EINVAL;
-    memcpy(&_program->num_counters, child->data,
-           sizeof(_program->num_counters));
-
     if (!(child = bf_marsh_next_child(marsh, child)))
         return -EINVAL;
     _program->img = bf_memdup(child->data, child->data_len);
@@ -368,7 +408,6 @@ void bf_program_dump(const struct bf_program *program, prefix_t *prefix)
 
     bf_dump_prefix_push(prefix);
 
-    DUMP(prefix, "num_counters: %lu", program->num_counters);
     DUMP(prefix, "prog_name: %s", program->prog_name);
 
     DUMP(prefix, "cmap: struct bf_map *");
@@ -381,6 +420,11 @@ void bf_program_dump(const struct bf_program *program, prefix_t *prefix)
     bf_map_dump(program->pmap, bf_dump_prefix_last(prefix));
     bf_dump_prefix_pop(prefix);
 
+    DUMP(prefix, "lmap: struct bf_map *");
+    bf_dump_prefix_push(prefix);
+    bf_map_dump(program->lmap, bf_dump_prefix_last(prefix));
+    bf_dump_prefix_pop(prefix);
+
     DUMP(prefix, "sets: bf_list<bf_map>[%lu]", bf_list_size(&program->sets));
     bf_dump_prefix_push(prefix);
     bf_list_foreach (&program->sets, map_node) {
@@ -429,18 +473,6 @@ void bf_program_dump(const struct bf_program *program, prefix_t *prefix)
     bf_dump_prefix_pop(prefix);
 }
 
-static inline size_t _bf_round_next_power_of_2(size_t value)
-{
-    value--;
-    value |= value >> 1;
-    value |= value >> 2;
-    value |= value >> 4;
-    value |= value >> 8;
-    value |= value >> 16;
-
-    return ++value;
-}
-
 int bf_program_grow_img(struct bf_program *program)
 {
     size_t new_cap = _BF_PROGRAM_DEFAULT_IMG_SIZE;
@@ -513,6 +545,10 @@ static int _bf_program_fixup(struct bf_program *program,
             insn_type = BF_FIXUP_INSN_IMM;
             value = program->pmap->fd;
             break;
+        case BF_FIXUP_TYPE_LOG_MAP_FD:
+            insn_type = BF_FIXUP_INSN_IMM;
+            value = program->lmap->fd;
+            break;
         case BF_FIXUP_TYPE_SET_MAP_FD:
             map = bf_list_get_at(&program->sets, fixup->attr.set_index);
             if (!map) {
@@ -522,11 +558,12 @@ static int _bf_program_fixup(struct bf_program *program,
             insn_type = BF_FIXUP_INSN_IMM;
             value = map->fd;
             break;
-        case BF_FIXUP_TYPE_FUNC_CALL:
+        case BF_FIXUP_ELFSTUB_CALL:
             insn_type = BF_FIXUP_INSN_IMM;
-            offset = program->functions_location[fixup->attr.function] -
+            offset = program->elfstubs_location[fixup->attr.elfstub_id] -
                      fixup->insn - 1;
-            bf_assert(offset < INT_MAX);
+            if (offset >= INT_MAX)
+                return bf_err_r(-EINVAL, "invalid ELF stub call offset");
             value = (int32_t)offset;
             break;
         default:
@@ -554,9 +591,10 @@ static int _bf_program_generate_rule(struct bf_program *program,
         struct bf_matcher *matcher = bf_list_node_get_data(matcher_node);
 
         switch (matcher->type) {
-        case BF_MATCHER_META_IFINDEX:
+        case BF_MATCHER_META_IFACE:
         case BF_MATCHER_META_L3_PROTO:
         case BF_MATCHER_META_L4_PROTO:
+        case BF_MATCHER_META_PROBABILITY:
         case BF_MATCHER_META_SPORT:
         case BF_MATCHER_META_DPORT:
             r = bf_matcher_generate_meta(program, matcher);
@@ -564,14 +602,19 @@ static int _bf_program_generate_rule(struct bf_program *program,
                 return r;
             break;
         case BF_MATCHER_IP4_SADDR:
+        case BF_MATCHER_IP4_SNET:
         case BF_MATCHER_IP4_DADDR:
+        case BF_MATCHER_IP4_DNET:
         case BF_MATCHER_IP4_PROTO:
             r = bf_matcher_generate_ip4(program, matcher);
             if (r)
                 return r;
             break;
         case BF_MATCHER_IP6_SADDR:
+        case BF_MATCHER_IP6_SNET:
         case BF_MATCHER_IP6_DADDR:
+        case BF_MATCHER_IP6_DNET:
+        case BF_MATCHER_IP6_NEXTHDR:
             r = bf_matcher_generate_ip6(program, matcher);
             if (r)
                 return r;
@@ -595,16 +638,36 @@ static int _bf_program_generate_rule(struct bf_program *program,
             if (r)
                 return r;
             break;
+        case BF_MATCHER_ICMP_TYPE:
+        case BF_MATCHER_ICMP_CODE:
+        case BF_MATCHER_ICMPV6_TYPE:
+        case BF_MATCHER_ICMPV6_CODE:
+            r = bf_matcher_generate_icmp(program, matcher);
+            if (r)
+                return r;
+            break;
         default:
             return bf_err_r(-EINVAL, "unknown matcher type %d", matcher->type);
         };
     }
 
+    if (rule->log) {
+        EMIT(program, BPF_MOV64_REG(BPF_REG_1, BPF_REG_10));
+        EMIT(program, BPF_ALU64_IMM(BPF_ADD, BPF_REG_1, BF_PROG_CTX_OFF(arg)));
+        EMIT_LOAD_LOG_FD_FIXUP(program, BPF_REG_2);
+        EMIT(program, BPF_MOV64_IMM(BPF_REG_3, rule->log));
+        EMIT(program, BPF_MOV64_REG(BPF_REG_4, BPF_REG_7));
+        EMIT(program, BPF_MOV64_REG(BPF_REG_5, BPF_REG_8));
+
+        EMIT_FIXUP_ELFSTUB(program, BF_ELFSTUB_LOG);
+    }
+
     if (rule->counters) {
-        EMIT(program, BPF_MOV32_IMM(BPF_REG_1, rule->index));
-        EMIT(program, BPF_LDX_MEM(BPF_DW, BPF_REG_2, BPF_REG_10,
-                                  BF_PROG_CTX_OFF(pkt_size)));
-        EMIT_FIXUP_CALL(program, BF_FIXUP_FUNC_UPDATE_COUNTERS);
+        EMIT(program, BPF_MOV64_REG(BPF_REG_1, BPF_REG_10));
+        EMIT(program, BPF_ALU64_IMM(BPF_ADD, BPF_REG_1, BF_PROG_CTX_OFF(arg)));
+        EMIT_LOAD_COUNTERS_FD_FIXUP(program, BPF_REG_2);
+        EMIT(program, BPF_MOV32_IMM(BPF_REG_3, rule->index));
+        EMIT_FIXUP_ELFSTUB(program, BF_ELFSTUB_UPDATE_COUNTERS);
     }
 
     switch (rule->verdict) {
@@ -631,72 +694,10 @@ static int _bf_program_generate_rule(struct bf_program *program,
     return 0;
 }
 
-/**
- * Generate the BPF function to update a rule's counters.
- *
- * This function defines a new function **in** the generated BPF program to
- * be called during packet processing.
- *
- * Parameters:
- * - @c r1 : index of the rule to update the counters for.
- * - @c r2 : size of the packet.
- * Returns:
- * 0 on success, non-zero on error.
- *
- * @param program Program to emit the function into. Can not be NULL.
- * @return 0 on success, or negative errno value on error.
- */
-static int _bf_program_generate_update_counters(struct bf_program *program)
-{
-    // Move the counters key in scratch[0..4] and the packet size in scratch[8..15]
-    EMIT(program,
-         BPF_STX_MEM(BPF_W, BPF_REG_10, BPF_REG_1, BF_PROG_SCR_OFF(0)));
-    EMIT(program,
-         BPF_STX_MEM(BPF_DW, BPF_REG_10, BPF_REG_2, BF_PROG_SCR_OFF(8)));
-
-    // Call bpf_map_lookup_elem()
-    EMIT_LOAD_COUNTERS_FD_FIXUP(program, BPF_REG_1);
-    EMIT(program, BPF_MOV64_REG(BPF_REG_2, BPF_REG_10));
-    EMIT(program, BPF_ALU64_IMM(BPF_ADD, BPF_REG_2, BF_PROG_SCR_OFF(0)));
-    EMIT(program, BPF_EMIT_CALL(BPF_FUNC_map_lookup_elem));
-
-    // If the counters doesn't exist, return from the function
-    {
-        _cleanup_bf_jmpctx_ struct bf_jmpctx _ =
-            bf_jmpctx_get(program, BPF_JMP_IMM(BPF_JNE, BPF_REG_0, 0, 0));
-
-        if (bf_opts_is_verbose(BF_VERBOSE_BPF))
-            EMIT_PRINT(program, "failed to fetch the rule's counters");
-
-        EMIT(program, BPF_MOV32_IMM(BPF_REG_0, 1));
-        EMIT(program, BPF_EXIT_INSN());
-    }
-
-    // Increment the packets count by 1.
-    EMIT(program, BPF_LDX_MEM(BPF_DW, BPF_REG_1, BPF_REG_0,
-                              offsetof(struct bf_counter, packets)));
-    EMIT(program, BPF_ALU64_IMM(BPF_ADD, BPF_REG_1, 1));
-    EMIT(program, BPF_STX_MEM(BPF_DW, BPF_REG_0, BPF_REG_1,
-                              offsetof(struct bf_counter, packets)));
-
-    // Increase the total byte by the size of the packet.
-    EMIT(program, BPF_LDX_MEM(BPF_DW, BPF_REG_1, BPF_REG_0,
-                              offsetof(struct bf_counter, bytes)));
-    EMIT(program,
-         BPF_LDX_MEM(BPF_DW, BPF_REG_2, BPF_REG_10, BF_PROG_SCR_OFF(8)));
-    EMIT(program, BPF_ALU64_REG(BPF_ADD, BPF_REG_1, BPF_REG_2));
-    EMIT(program, BPF_STX_MEM(BPF_DW, BPF_REG_0, BPF_REG_1,
-                              offsetof(struct bf_counter, bytes)));
-
-    // On success, return 0
-    EMIT(program, BPF_MOV32_IMM(BPF_REG_0, 0));
-    EMIT(program, BPF_EXIT_INSN());
-
-    return 0;
-}
-
-static int _bf_program_generate_functions(struct bf_program *program)
+static int _bf_program_generate_elfstubs(struct bf_program *program)
 {
+    const struct bf_elfstub *elfstub;
+    size_t start_at;
     int r;
 
     bf_assert(program);
@@ -705,29 +706,58 @@ static int _bf_program_generate_functions(struct bf_program *program)
         struct bf_fixup *fixup = bf_list_node_get_data(fixup_node);
         size_t off = program->img_size;
 
-        if (fixup->type != BF_FIXUP_TYPE_FUNC_CALL)
+        if (fixup->type != BF_FIXUP_ELFSTUB_CALL)
             continue;
 
-        bf_assert(fixup->attr.function >= 0 &&
-                  fixup->attr.function < _BF_FIXUP_FUNC_MAX);
-
-        // Only generate each function once
-        if (program->functions_location[fixup->attr.function])
+        // Only generate each ELF stub once
+        if (program->elfstubs_location[fixup->attr.elfstub_id])
             continue;
 
-        switch (fixup->attr.function) {
-        case BF_FIXUP_FUNC_UPDATE_COUNTERS:
-            r = _bf_program_generate_update_counters(program);
-            if (r)
-                return r;
-            break;
-        default:
-            bf_abort("unsupported fixup function, this should not happen: %d",
-                     fixup->attr.function);
-            break;
+        bf_dbg("generate ELF stub for ID %d", fixup->attr.elfstub_id);
+
+        elfstub = bf_ctx_get_elfstub(fixup->attr.elfstub_id);
+        if (!elfstub) {
+            return bf_err_r(-ENOENT, "no ELF stub found for ID %d",
+                            fixup->attr.elfstub_id);
         }
 
-        program->functions_location[fixup->attr.function] = off;
+        start_at = program->img_size;
+
+        for (size_t i = 0; i < elfstub->ninsns; ++i) {
+            r = bf_program_emit(program, elfstub->insns[i]);
+            if (r)
+                return bf_err_r(r, "failed to insert ELF stub instruction");
+        }
+
+        bf_list_foreach (&elfstub->strs, pstr_node) {
+            _free_bf_fixup_ struct bf_fixup *fixup = NULL;
+            struct bf_printk_str *pstr = bf_list_node_get_data(pstr_node);
+            size_t insn_idx = start_at + pstr->insn_idx;
+            const struct bf_printer_msg *msg =
+                bf_printer_add_msg(program->printer, pstr->str);
+            struct bpf_insn ld_insn[2] = {
+                BPF_LD_MAP_FD(BPF_REG_1, 0),
+            };
+
+            ld_insn[0].src_reg = BPF_PSEUDO_MAP_VALUE;
+            ld_insn[1].imm = (int)bf_printer_msg_offset(msg);
+
+            program->img[insn_idx] = ld_insn[0];
+            program->img[insn_idx + 1] = ld_insn[1];
+
+            r = bf_fixup_new(&fixup, BF_FIXUP_TYPE_PRINTER_MAP_FD, insn_idx,
+                             NULL);
+            if (r)
+                return r;
+
+            r = bf_list_add_tail(&program->fixups, fixup);
+            if (r)
+                return r;
+
+            TAKE_PTR(fixup);
+        }
+
+        program->elfstubs_location[fixup->attr.elfstub_id] = off;
     }
 
     return 0;
@@ -773,7 +803,7 @@ int bf_program_emit_kfunc_call(struct bf_program *program, const char *name)
 int bf_program_emit_fixup(struct bf_program *program, enum bf_fixup_type type,
                           struct bpf_insn insn, const union bf_fixup_attr *attr)
 {
-    _cleanup_bf_fixup_ struct bf_fixup *fixup = NULL;
+    _free_bf_fixup_ struct bf_fixup *fixup = NULL;
     int r;
 
     bf_assert(program);
@@ -803,10 +833,10 @@ int bf_program_emit_fixup(struct bf_program *program, enum bf_fixup_type type,
     return 0;
 }
 
-int bf_program_emit_fixup_call(struct bf_program *program,
-                               enum bf_fixup_func function)
+int bf_program_emit_fixup_elfstub(struct bf_program *program,
+                                  enum bf_elfstub_id id)
 {
-    _cleanup_bf_fixup_ struct bf_fixup *fixup = NULL;
+    _free_bf_fixup_ struct bf_fixup *fixup = NULL;
     int r;
 
     bf_assert(program);
@@ -817,11 +847,11 @@ int bf_program_emit_fixup_call(struct bf_program *program,
             return r;
     }
 
-    r = bf_fixup_new(&fixup, BF_FIXUP_TYPE_FUNC_CALL, program->img_size, NULL);
+    r = bf_fixup_new(&fixup, BF_FIXUP_ELFSTUB_CALL, program->img_size, NULL);
     if (r)
         return r;
 
-    fixup->attr.function = function;
+    fixup->attr.elfstub_id = id;
 
     r = bf_list_add_tail(&program->fixups, fixup);
     if (r)
@@ -829,10 +859,6 @@ int bf_program_emit_fixup_call(struct bf_program *program,
 
     TAKE_PTR(fixup);
 
-    /* This call could fail and return an error, in which case it is not
-     * properly handled. However, this shouldn't be an issue as we previously
-     * test whether enough room is available in cgen.img, which is currently
-     * the only reason for EMIT() to fail. */
     EMIT(program, BPF_CALL_REL(0));
 
     return 0;
@@ -843,11 +869,6 @@ int bf_program_generate(struct bf_program *program)
     const struct bf_chain *chain = program->runtime.chain;
     int r;
 
-    /* Add 1 to the number of counters for the policy counter, and 1
-     * for the first reserved error slot. This must be done ahead of
-     * generation, as we will index into the error counters. */
-    program->num_counters = bf_list_size(&chain->rules) + 2;
-
     // Save the program's argument into the context.
     EMIT(program,
          BPF_STX_MEM(BPF_DW, BPF_REG_10, BPF_REG_1, BF_PROG_CTX_OFF(arg)));
@@ -856,6 +877,21 @@ int bf_program_generate(struct bf_program *program)
     EMIT(program, BPF_MOV64_IMM(BPF_REG_7, 0));
     EMIT(program, BPF_MOV64_IMM(BPF_REG_8, 0));
 
+    program->ipv6_nexthdr = false;
+    bf_list_foreach (&chain->rules, rule_node) {
+        struct bf_rule *rule = bf_list_node_get_data(rule_node);
+        bf_list_foreach (&rule->matchers, matcher_node) {
+            struct bf_matcher *matcher = bf_list_node_get_data(matcher_node);
+            if (matcher->type == BF_MATCHER_IP6_NEXTHDR) {
+                // Zeroing IPv6 extension headers
+                EMIT(program, BPF_STX_MEM(BPF_DW, BPF_REG_10, BPF_REG_7,
+                                          BF_PROG_CTX_OFF(ipv6_eh)));
+                program->ipv6_nexthdr = true;
+                break;
+            }
+        }
+    }
+
     r = program->runtime.ops->gen_inline_prologue(program);
     if (r)
         return r;
@@ -872,22 +908,24 @@ int bf_program_generate(struct bf_program *program)
         return r;
 
     // Call the update counters function
-    EMIT(program, BPF_MOV32_IMM(BPF_REG_1, bf_list_size(&chain->rules)));
+    EMIT(program, BPF_MOV64_REG(BPF_REG_1, BPF_REG_10));
+    EMIT(program, BPF_ALU64_IMM(BPF_ADD, BPF_REG_1, BF_PROG_CTX_OFF(arg)));
+    EMIT_LOAD_COUNTERS_FD_FIXUP(program, BPF_REG_2);
     EMIT(program,
-         BPF_LDX_MEM(BPF_DW, BPF_REG_2, BPF_REG_10, BF_PROG_CTX_OFF(pkt_size)));
-    EMIT_FIXUP_CALL(program, BF_FIXUP_FUNC_UPDATE_COUNTERS);
+         BPF_MOV32_IMM(BPF_REG_3, bf_program_chain_counter_idx(program)));
+    EMIT_FIXUP_ELFSTUB(program, BF_ELFSTUB_UPDATE_COUNTERS);
 
     EMIT(program, BPF_MOV64_IMM(BPF_REG_0, program->runtime.ops->get_verdict(
                                                chain->policy)));
     EMIT(program, BPF_EXIT_INSN());
 
-    r = _bf_program_generate_functions(program);
+    r = _bf_program_generate_elfstubs(program);
     if (r)
         return r;
 
-    r = _bf_program_fixup(program, BF_FIXUP_TYPE_FUNC_CALL);
+    r = _bf_program_fixup(program, BF_FIXUP_ELFSTUB_CALL);
     if (r)
-        return bf_err_r(r, "failed to generate function call fixups");
+        return bf_err_r(r, "failed to generate ELF stub call fixups");
 
     return 0;
 }
@@ -933,7 +971,8 @@ static int _bf_program_load_counters_map(struct bf_program *program)
 
     bf_assert(program);
 
-    r = bf_map_set_n_elems(program->cmap, program->num_counters);
+    r = bf_map_set_n_elems(program->cmap,
+                           bf_list_size(&program->runtime.chain->rules) + 2);
     if (r < 0)
         return r;
 
@@ -950,6 +989,26 @@ static int _bf_program_load_counters_map(struct bf_program *program)
     return 0;
 }
 
+static int _bf_program_load_log_map(struct bf_program *program)
+{
+    _cleanup_close_ int _fd = -1;
+    int r;
+
+    bf_assert(program);
+
+    r = bf_map_create(program->lmap, 0);
+    if (r < 0)
+        return r;
+
+    r = _bf_program_fixup(program, BF_FIXUP_TYPE_LOG_MAP_FD);
+    if (r < 0) {
+        bf_map_destroy(program->lmap);
+        return bf_err_r(r, "failed to fixup log map FD");
+    }
+
+    return 0;
+}
+
 static int _bf_program_load_sets_maps(struct bf_program *new_prog)
 {
     const bf_list_node *set_node;
@@ -1044,6 +1103,10 @@ int bf_program_load(struct bf_program *prog)
     if (r)
         return r;
 
+    r = _bf_program_load_log_map(prog);
+    if (r)
+        return r;
+
     if (bf_opts_is_verbose(BF_VERBOSE_DEBUG)) {
         log_buf = malloc(_BF_LOG_BUF_SIZE);
         if (!log_buf) {
@@ -1052,6 +1115,9 @@ int bf_program_load(struct bf_program *prog)
         }
     }
 
+    if (bf_opts_is_verbose(BF_VERBOSE_BYTECODE))
+        bf_program_dump_bytecode(prog);
+
     r = bf_bpf_prog_load(
         prog->prog_name, bf_hook_to_bpf_prog_type(prog->runtime.chain->hook),
         prog->img, prog->img_size,
@@ -1062,9 +1128,6 @@ int bf_program_load(struct bf_program *prog)
                         prog->img_size, log_buf ? log_buf : "<NO LOG BUFFER>");
     }
 
-    if (bf_opts_is_verbose(BF_VERBOSE_BYTECODE))
-        bf_program_dump_bytecode(prog);
-
     return r;
 }
 
@@ -1099,6 +1162,7 @@ void bf_program_unload(struct bf_program *prog)
     bf_link_detach(prog->link);
     bf_map_destroy(prog->cmap);
     bf_map_destroy(prog->pmap);
+    bf_map_destroy(prog->lmap);
     bf_list_foreach (&prog->sets, map_node)
         bf_map_destroy(bf_list_node_get_data(map_node));
 }
@@ -1154,6 +1218,12 @@ int bf_program_pin(struct bf_program *prog, int dir_fd)
         goto err_unpin_all;
     }
 
+    r = bf_map_pin(prog->lmap, dir_fd);
+    if (r) {
+        bf_err_r(r, "failed to pin BPF log map for '%s'", name);
+        goto err_unpin_all;
+    }
+
     bf_list_foreach (&prog->sets, set_node) {
         r = bf_map_pin(bf_list_node_get_data(set_node), dir_fd);
         if (r) {
@@ -1184,6 +1254,7 @@ void bf_program_unpin(struct bf_program *prog, int dir_fd)
 
     bf_map_unpin(prog->cmap, dir_fd);
     bf_map_unpin(prog->pmap, dir_fd);
+    bf_map_unpin(prog->lmap, dir_fd);
 
     bf_list_foreach (&prog->sets, set_node)
         bf_map_unpin(bf_list_node_get_data(set_node), dir_fd);
@@ -1192,3 +1263,13 @@ void bf_program_unpin(struct bf_program *prog, int dir_fd)
 
     unlinkat(dir_fd, prog->prog_name, 0);
 }
+
+size_t bf_program_chain_counter_idx(const struct bf_program *program)
+{
+    return bf_list_size(&program->runtime.chain->rules);
+}
+
+size_t bf_program_error_counter_idx(const struct bf_program *program)
+{
+    return bf_list_size(&program->runtime.chain->rules) + 1;
+}
