@@ -4,10 +4,10 @@
  */
 
 #include <errno.h>
-#include <stdlib.h>
 
 #include "bpfilter/cgen/cgen.h"
 #include "bpfilter/cgen/prog/link.h"
+#include "bpfilter/cgen/prog/map.h"
 #include "bpfilter/cgen/program.h"
 #include "bpfilter/ctx.h"
 #include "bpfilter/xlate/front.h"
@@ -16,6 +16,7 @@
 #include "core/front.h"
 #include "core/helper.h"
 #include "core/hook.h"
+#include "core/io.h"
 #include "core/list.h"
 #include "core/logger.h"
 #include "core/marsh.h"
@@ -62,10 +63,10 @@ int _bf_cli_ruleset_flush(const struct bf_request *request,
 static int _bf_cli_ruleset_get(const struct bf_request *request,
                                struct bf_response **response)
 {
-    _cleanup_bf_marsh_ struct bf_marsh *marsh = NULL;
-    _cleanup_bf_marsh_ struct bf_marsh *chain_marsh = NULL;
-    _cleanup_bf_marsh_ struct bf_marsh *hookopts_marsh = NULL;
-    _cleanup_bf_marsh_ struct bf_marsh *counters_marsh = NULL;
+    _free_bf_marsh_ struct bf_marsh *marsh = NULL;
+    _free_bf_marsh_ struct bf_marsh *chain_marsh = NULL;
+    _free_bf_marsh_ struct bf_marsh *hookopts_marsh = NULL;
+    _free_bf_marsh_ struct bf_marsh *counters_marsh = NULL;
     _clean_bf_list_ bf_list cgens = bf_list_default(NULL, NULL);
     _clean_bf_list_ bf_list chains = bf_list_default(NULL, bf_chain_marsh);
     _clean_bf_list_ bf_list hookopts = bf_list_default(NULL, bf_hookopts_marsh);
@@ -85,7 +86,7 @@ static int _bf_cli_ruleset_get(const struct bf_request *request,
 
     bf_list_foreach (&cgens, cgen_node) {
         struct bf_cgen *cgen = bf_list_node_get_data(cgen_node);
-        _cleanup_bf_list_ bf_list *cgen_counters = NULL;
+        _free_bf_list_ bf_list *cgen_counters = NULL;
 
         r = bf_list_add_tail(&chains, cgen->chain);
         if (r)
@@ -124,7 +125,7 @@ static int _bf_cli_ruleset_get(const struct bf_request *request,
     r = bf_marsh_new(&hookopts_marsh, NULL, 0);
     bf_list_foreach (&hookopts, hookopts_node) {
         struct bf_hookopts *hookopts = bf_list_node_get_data(hookopts_node);
-        _cleanup_bf_marsh_ struct bf_marsh *child = NULL;
+        _free_bf_marsh_ struct bf_marsh *child = NULL;
 
         if (hookopts) {
             r = bf_hookopts_marsh(bf_list_node_get_data(hookopts_node), &child);
@@ -178,8 +179,8 @@ int _bf_cli_ruleset_set(const struct bf_request *request,
     bf_ctx_flush(BF_FRONT_CLI);
 
     while ((list_elem = bf_marsh_next_child(marsh, list_elem))) {
-        _cleanup_bf_cgen_ struct bf_cgen *cgen = NULL;
-        _cleanup_bf_chain_ struct bf_chain *chain = NULL;
+        _free_bf_cgen_ struct bf_cgen *cgen = NULL;
+        _free_bf_chain_ struct bf_chain *chain = NULL;
         _free_bf_hookopts_ struct bf_hookopts *hookopts = NULL;
         struct bf_marsh *child = NULL;
 
@@ -234,8 +235,8 @@ int _bf_cli_chain_set(const struct bf_request *request,
 {
     struct bf_cgen *old_cgen;
     struct bf_marsh *marsh, *child = NULL;
-    _cleanup_bf_cgen_ struct bf_cgen *new_cgen = NULL;
-    _cleanup_bf_chain_ struct bf_chain *chain = NULL;
+    _free_bf_cgen_ struct bf_cgen *new_cgen = NULL;
+    _free_bf_chain_ struct bf_chain *chain = NULL;
     _free_bf_hookopts_ struct bf_hookopts *hookopts = NULL;
     int r;
 
@@ -293,7 +294,7 @@ int _bf_cli_chain_set(const struct bf_request *request,
 static int _bf_cli_chain_get(const struct bf_request *request,
                              struct bf_response **response)
 {
-    _cleanup_bf_marsh_ struct bf_marsh *marsh = NULL;
+    _free_bf_marsh_ struct bf_marsh *marsh = NULL;
     _clean_bf_list_ bf_list counters =
         bf_list_default(bf_counter_free, bf_counter_marsh);
     struct bf_cgen *cgen;
@@ -325,7 +326,7 @@ static int _bf_cli_chain_get(const struct bf_request *request,
         return bf_err_r(r, "failed to get new marsh");
 
     {
-        _cleanup_bf_marsh_ struct bf_marsh *child = NULL;
+        _free_bf_marsh_ struct bf_marsh *child = NULL;
 
         r = bf_chain_marsh(cgen->chain, &child);
         if (r)
@@ -337,7 +338,7 @@ static int _bf_cli_chain_get(const struct bf_request *request,
     }
 
     if (cgen->program->link->hookopts) {
-        _cleanup_bf_marsh_ struct bf_marsh *child = NULL;
+        _free_bf_marsh_ struct bf_marsh *child = NULL;
 
         r = bf_hookopts_marsh(cgen->program->link->hookopts, &child);
         if (r)
@@ -353,7 +354,7 @@ static int _bf_cli_chain_get(const struct bf_request *request,
     }
 
     {
-        _cleanup_bf_marsh_ struct bf_marsh *child = NULL;
+        _free_bf_marsh_ struct bf_marsh *child = NULL;
 
         r = bf_cgen_get_counters(cgen, &counters);
         if (r)
@@ -372,12 +373,51 @@ static int _bf_cli_chain_get(const struct bf_request *request,
                                    bf_marsh_size(marsh));
 }
 
+int _bf_cli_chain_logs_fd(const struct bf_request *request,
+                          struct bf_response **response)
+{
+    struct bf_cgen *cgen;
+    struct bf_marsh *marsh, *child = NULL;
+    int r;
+
+    UNUSED(response);
+
+    marsh = (struct bf_marsh *)request->data;
+    if (bf_marsh_size(marsh) != request->data_len) {
+        return bf_err_r(
+            -EINVAL,
+            "request payload is expected to have the same size as the marsh");
+    }
+
+    if (!(child = bf_marsh_next_child(marsh, child)))
+        return -EINVAL;
+    if (child->data_len < 2)
+        return bf_err_r(-EINVAL, "_bf_cli_chain_logs_fd: chain name is empty");
+    if (child->data[child->data_len - 1]) {
+        return bf_err_r(
+            -EINVAL, "_bf_cli_chain_logs_fd: chain name if not nul-terminated");
+    }
+
+    cgen = bf_ctx_get_cgen(child->data);
+    if (!cgen)
+        return bf_err_r(-ENOENT, "failed to find chain '%s'", child->data);
+
+    if (!cgen->program || !cgen->program->lmap->fd)
+        return bf_err_r(-ENOENT, "chain '%s' has no logs buffer", child->data);
+
+    r = bf_send_fd(request->fd, cgen->program->lmap->fd);
+    if (r < 0)
+        return bf_err_r(errno, "failed to send logs FD for '%s'", child->data);
+
+    return 0;
+}
+
 int _bf_cli_chain_load(const struct bf_request *request,
                        struct bf_response **response)
 {
     struct bf_marsh *marsh, *child = NULL;
-    _cleanup_bf_cgen_ struct bf_cgen *cgen = NULL;
-    _cleanup_bf_chain_ struct bf_chain *chain = NULL;
+    _free_bf_cgen_ struct bf_cgen *cgen = NULL;
+    _free_bf_chain_ struct bf_chain *chain = NULL;
     int r;
 
     bf_assert(request && response);
@@ -427,7 +467,7 @@ int _bf_cli_chain_attach(const struct bf_request *request,
                          struct bf_response **response)
 {
     struct bf_marsh *marsh, *child = NULL;
-    _cleanup_bf_chain_ struct bf_chain *chain = NULL;
+    _free_bf_chain_ struct bf_chain *chain = NULL;
     struct bf_cgen *cgen = NULL;
     const char *name;
     _free_bf_hookopts_ struct bf_hookopts *hookopts = NULL;
@@ -477,7 +517,7 @@ int _bf_cli_chain_update(const struct bf_request *request,
                          struct bf_response **response)
 {
     struct bf_marsh *marsh, *child = NULL;
-    _cleanup_bf_chain_ struct bf_chain *chain = NULL;
+    _free_bf_chain_ struct bf_chain *chain = NULL;
     struct bf_cgen *cgen = NULL;
     int r;
 
@@ -567,6 +607,9 @@ static int _bf_cli_request_handler(struct bf_request *request,
     case BF_REQ_CHAIN_GET:
         r = _bf_cli_chain_get(request, response);
         break;
+    case BF_REQ_CHAIN_LOGS_FD:
+        r = _bf_cli_chain_logs_fd(request, response);
+        break;
     case BF_REQ_CHAIN_LOAD:
         r = _bf_cli_chain_load(request, response);
         break;
